import warnings
from typing import Optional

from rapidfuzz.process import extractOne

from lol_id_tools.local_data_parser import get_clean_locale
from lol_id_tools.logger import lit_logger
from lol_id_tools.lol_object_data import LolObjectData

# Instantiating a LolObjectData object is very light as all its fields are ghost loaded.
lod = LolObjectData()


# TODO Add older patches support
def get_name(
    input_id: int, output_locale: str = "en_US", object_type=None, retry=True, fallback_to_none=True
) -> Optional[str]:
    """Gets you the name of the associated Riot object.

    If the chosen locale is not loaded in the database, it will be loaded before returning the result.

    Args:
        input_id: Riot ID of the object
        output_locale: Locale of the output
        object_type: specifics the object type you want the name of, in ['champion', 'item', 'rune', 'summoner_spell']
        retry: Optional variable specifying if local_data should be reloaded once if the object cannot be found
        fallback_to_none: whether the tool returns None in case of a KeyError

    Returns:
        The matching object name.

    Raises:
        KeyError: The corresponding object was not found.
        ValueError: The locale was not understood properly.

    Usage example:
        get_name(21)
        get_name(21, 'ko_KR')
    """
    try:
        input_id = int(input_id)
    except ValueError:
        raise ValueError(f"{input_id} could not be cast to an integer.")

    # Riot uses 0 as a "no item" value and -1 as "no ban" value.
    if input_id <= 0:
        return ""

    # We start by cleaning up our locale
    output_locale = get_clean_locale(output_locale)

    # First, we see if the object is there with the given constraints
    try:
        if not object_type:
            if lod.loaded_data[output_locale][input_id].__len__() > 1:
                warning_text = f"Multiple objects with ID {input_id} found, please inform object_type."
                lit_logger.warning(warning_text)
            for object_type in ["champion", "item", "rune", "summoner_spell"]:
                # Iterating this way to have a priority between object types
                # TODO Rework that for more readable code
                if object_type in lod.loaded_data[output_locale][input_id]:
                    break

        return lod.loaded_data[output_locale][input_id][object_type]
    except KeyError:
        pass

    if output_locale not in lod.loaded_locales:
        lod.load_locale(output_locale)
        return get_name(input_id, output_locale, retry=False)

    if retry:
        lod.reload_all_locales()
        return get_name(input_id, output_locale, retry=False)

    error_text = f"Riot object with id {input_id} could not be found."

    if not fallback_to_none:
        raise KeyError(error_text)
    else:
        lit_logger.warning(error_text)
        return None


class NoMatchingNameFound(Exception):
    pass


def get_id(
    input_str: str,
    minimum_score: int = 75,
    input_locale: str = None,
    object_type: str = None,
    retry: bool = True,
) -> int:
    """Returns the best Riot ID guess for the given name.

    Will return the id that matches the query best. Best used in curated environments.

    Args:
        input_str: Search string.
        minimum_score: Optional minimum ratio (between 0 and 100) under which the function retries or raises.
        input_locale: The language the input was in.
        object_type: Optional string in ['champion', 'item', 'rune', 'summoner_spell']
        retry: Optional variable specifying if local_data should be reloaded if the object cannot be found.

    Returns:
        The matching object ID.

    Raises:
        NoMatchingNameFound: No string scored above the minimum_score.
        ValueError: The locale was not understood properly.

    Usage example:
        get_id('Miss Fortune')
        get_id('MF')
        get_id('미스 포츈')
        get_id('Dio')
    """
    input_str = input_str.lower()

    # Handling some Leaguepedia special cases as having an ID of 0, might be stupid and should just raise
    if not input_str or input_str == "none" or input_str == "loss of ban" or input_str == "no item":
        return 0

    # We try to directly get the object with the exact input name
    try:
        return lod.names_to_id[input_str].id
    except KeyError:
        pass

    # If we run get_id() with no locale and nothing is loaded, we load english by default.
    if not input_locale and not lod.loaded_locales:
        lod.load_locale("en_US")

    possible_names_to_id = lod.names_to_id

    if object_type:
        possible_names_to_id = {
            name: possible_names_to_id[name]
            for name in possible_names_to_id
            if possible_names_to_id[name].object_type == object_type
        }

    if input_locale:
        locale = get_clean_locale(input_locale)

        # If the locale was not loaded yet, we restart the process
        if locale not in lod.loaded_locales:
            lod.load_locale(locale)
            return get_id(input_str, minimum_score, input_locale, object_type, retry=False)

        possible_names_to_id = {
            name: possible_names_to_id[name]
            for name in possible_names_to_id
            if possible_names_to_id[name].locale == locale
        }

    name_guess, score, idx = extractOne(input_str, possible_names_to_id.keys())

    # TODO Handle multiple objects having the same score (happens with substrings of longer names)
    lit_logger.info(f"Name guess was {name_guess} from {input_str}")

    if score < minimum_score:
        if retry:
            lod.reload_all_locales()
            return get_id(input_str, minimum_score, input_locale, object_type, retry=False)
        else:
            error_text = f"No object name close enough to '{input_str}' found."
            raise NoMatchingNameFound(error_text)

    return lod.names_to_id[name_guess].id


def get_translation(
    object_name: str,
    output_locale: str = "en_US",
    minimum_score: int = 75,
    input_locale: str = None,
    object_type: str = None,
    retry: bool = True,
) -> str:
    """Returns the best translation guess for the given name.

    Will return the id that matches the query best. Best used in curated environments.

    Args:
        object_name: Search string
        output_locale: The language to translate to, with 'en_US' as the default value
        minimum_score: Optional minimum ratio (between 0 and 100) under which the function retries or raises
        input_locale: The language the input was in
        object_type: Optional string in ['champion', 'item', 'rune', 'summoner_spell']
        retry: Optional variable specifying if local_data should be reloaded if the object cannot be found

    Returns:
        The matching object ID

    Raises:
        NoResultFound: No result scored above the minimum_score
        ValueError: One of the input locales was not understood properly

    Usage example:
        get_translation('미스 포츈')
        get_translation('Miss Fortune', 'ko_KR')
        get_translation('mf')
    """
    return get_name(get_id(object_name, minimum_score, input_locale, object_type, retry), output_locale,)



from lol_id_tools.logger import lit_logger
import requests

dd_url = "https://ddragon.leagueoflegends.com"

# TODO Move data from ddragon to cdragon to get ID of removed items too


def load_riot_objects(local_data, latest_version, locale: str, object_type: str):
    """Loads the selected type of objects in the database.

    Queries the Riot servers for the local_data then writes it to the database.
    All three object types are meant to be called together.

    Args:
        local_data: Dictionary to write to
        latest_version: the latest version available on dd
        locale: the locale to load
        object_type: the type of object to load__
    """
    url = get_ddragon_url(latest_version, locale, object_type)

    lit_logger.debug(f"Querying {url}")
    response = requests.get(url)

    riot_data = response.json()

    if object_type == "champion":
        parse_champions(riot_data, locale, local_data)
    elif object_type == "item":
        parse_items(riot_data, locale, local_data)
    elif object_type == "runesReforged":
        parse_runes(riot_data, locale, local_data)
    elif object_type == "summoner":
        parse_summoner_spells(riot_data, locale, local_data)


def get_ddragon_url(latest_version, locale: str, object_type: str):
    # Riot changed name for some reason
    return f"{dd_url}/cdn/{latest_version}/data/{locale}/{object_type}.json"


def parse_champions(data, locale, local_data):
    for champion_tag, champion_dict in data["data"].items():
        local_data[locale][int(champion_dict["key"])]["champion"] = champion_dict["name"]


def parse_items(data, locale, local_data):
    for item_id, item_dict in data["data"].items():
        if "Enchantment" not in item_dict["name"]:
            local_data[locale][int(item_id)]["item"] = item_dict["name"]
        else:
            for from_item_id in item_dict["from"]:
                from_item = data["data"][from_item_id]
                if "Jungle" in from_item["tags"]:
                    local_data[locale][int(item_id)][
                        "item"
                    ] = f"{from_item['name']} - {item_dict['name'].replace('Enchantment: ', '')}"


def parse_runes(data, locale, local_data):
    for rune_tree in data:
        # Adding tree names
        local_data[locale][rune_tree["id"]]["rune"] = rune_tree["name"]
        for slot in rune_tree["slots"]:
            for rune in slot["runes"]:
                local_data[locale][rune["id"]]["rune"] = rune["name"]


def parse_summoner_spells(riot_data, locale, local_data):
    for summoner_spell in riot_data["data"]:
        summoner_spell_info = riot_data["data"][summoner_spell]
        local_data[locale][int(summoner_spell_info["key"])]["summoner_spell"] = summoner_spell_info["name"]


def parse_cdragon_runes(local_data, locale):
    # TODO Clean this up
    cdragon_locale = locale.lower() if locale != "en_US" else "default"
    url = f"http://raw.communitydragon.org/latest/plugins/rcp-be-lol-game-data/global/{cdragon_locale}/v1/perks.json"

    lit_logger.debug(f"Querying {url}")
    response = requests.get(url)
    cdragon_data = response.json()

    for rune in cdragon_data:
        # TODO Get perks only in a cleaner way
        if 5000 < rune["id"] < 5010:
            local_data[locale][rune["id"]]["rune"] = rune["name"]



import os
import requests
import pickle
import logging

from concurrent.futures.thread import ThreadPoolExecutor
from collections import defaultdict
from typing import Dict

from lol_id_tools.local_data_parser import load_nickname_data, NameInfo
from lol_id_tools.data_parser import load_riot_objects, parse_cdragon_runes

save_folder = os.path.join(os.path.expanduser("~"), ".config", "lol_id_tools")
if not os.path.exists(save_folder):
    os.makedirs(save_folder)


class LolObjectData:
    """A class handling data about LoL objects.

    Everything is class-wide to make sure multiple programs on the same machine use the same data.
    """

    data_location = os.path.join(save_folder, "loaded_data.pkl")

    # riot_data represents all the data that we got from Riot and is ghost loaded for module loading efficiency
    # it is used directly for id -> name matching
    # riot_data[locale][id][object_type][NameInfo]
    _loaded_data = None

    @property
    def loaded_data(self):
        if self._loaded_data is None:
            self._loaded_data = self.unpickle_loaded_data()
            self.recalculate_names_to_id()
        return self._loaded_data

    # Pickling it to minimise web requests
    def pickle_loaded_data(self):
        with open(self.data_location, "wb+") as file:
            pickle.dump(self.loaded_data, file)

    def unpickle_loaded_data(self) -> Dict[str, Dict[int, Dict[str, str]]]:
        try:
            with open(self.data_location, "rb") as file:
                return pickle.load(file)
        except FileNotFoundError:
            return {}

    # nicknames_data is a dictionary representation of /locale_data/nicknames.json
    # It is not meant to be used directly as names_to_id is what we use for name -> id
    # nickname_data[locale][nickname][clean_name]
    _nickname_data = None

    @property
    def nickname_data(self):
        if self._nickname_data is None:
            self._nickname_data = load_nickname_data()
        return self._nickname_data

    # names_to_id is a flipped dictionary that holds names from both riot_data and nickname_data
    # names_to_id[name][IdInfo]
    _names_to_id: Dict[str, NameInfo] = {}

    @property
    def names_to_id(self):
        if not self._names_to_id:
            self.recalculate_names_to_id()
        return self._names_to_id

    # get_id() relies a lot on item names, so we also define a reversed dict with lowercase names as property
    def recalculate_names_to_id(self):
        for locale in self.loaded_data:
            # First we write the info from riot_data
            for id_ in self.loaded_data[locale]:
                for object_type in self.loaded_data[locale][id_]:
                    name = self.loaded_data[locale][id_][object_type]
                    self._names_to_id[name.lower()] = NameInfo(id_, object_type, locale)

            # Then we write the info from nicknames_data if we have loaded the locale
            if locale in self.nickname_data:
                for nickname in self.nickname_data[locale]:
                    clean_name = self.nickname_data[locale][nickname]
                    object_id = None
                    object_type = None
                    for id_ in self.loaded_data[locale]:
                        for object_type in self.loaded_data[locale][id_]:
                            if self.loaded_data[locale][id_][object_type] == clean_name:
                                object_id = id_
                                break

                    self._names_to_id[nickname.lower()] = NameInfo(object_id, object_type, locale)

    # Defining another property for more readable code
    @property
    def loaded_locales(self):
        return [k for k in self.loaded_data]

    def load_locale(self, locale, latest_version=None):
        if not latest_version:
            latest_version = self.get_latest_version()

        self.loaded_data[locale] = defaultdict(dict)

        with ThreadPoolExecutor() as executor:
            # TODO Just call different functions?
            for object_type in ["champion", "runesReforged", "item", "summoner"]:
                executor.submit(load_riot_objects, self.loaded_data, latest_version, locale, object_type)

            # Cdragon is different enough that it’s handled by itself
            executor.submit(parse_cdragon_runes, self.loaded_data, locale)

        self.recalculate_names_to_id()
        self.pickle_loaded_data()

    def reload_all_locales(self):
        self._names_to_id = {}
        latest_version = self.get_latest_version()

        with ThreadPoolExecutor() as executor:
            for locale in self.loaded_locales:
                executor.submit(self.load_locale, locale, latest_version)

    @staticmethod
    def get_latest_version():
        """Gets the latest version available on ddragon.
        """
        url = "https://ddragon.leagueoflegends.com/api/versions.json"

        response = requests.get(url)
        logging.debug(f"Querying {url}")

        data = response.json()
        return data[0]

    def delete_local_data(self):
        """Mainly used for testing purposes
        """
        try:
            os.remove(self.data_location)
        except FileNotFoundError:
            pass



import json
import os
from collections import namedtuple
from typing import Dict

from rapidfuzz.process import extractOne

NameInfo = namedtuple("NameInfo", ["id", "object_type", "locale"])


def load_nickname_data() -> Dict[str, Dict[str, str]]:
    """Returns the parsed nicknames.json.
    """
    with open(os.path.join(os.path.dirname(__file__), "local_data", "nicknames.json"), encoding="utf-8") as file:
        return json.load(file)


with open(os.path.join(os.path.dirname(__file__), "local_data", "locales.json"), encoding="utf-8") as file:
    locales_dict = json.load(file)


def get_clean_locale(locale: str):
    """Returns a "clean" locale from a user-entered input.
    """
    # We check if the locale is "clean". If not, we match it to our language list.
    if locale in locales_dict.values():
        return locale

    matching_language, score, idx = extractOne(locale, locales_dict.keys())
    if score > 80:
        return locales_dict[matching_language]
    else:
        raise ValueError("The locale name could not be understood")



import setuptools
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setuptools.setup(
    name="lol_id_tools",
    version="1.7.1",
    packages=["lol_id_tools", "lol_id_tools/local_data"],
    package_data={"": ["*.json"]},
    url="https://github.com/mrtolkien/lol_id_tools",
    license="MIT",
    author="Tolki",
    install_requires=["rapidfuzz>=0.14", "requests"],
    author_email="gary.mialaret+pypi@gmail.com",
    description="An id tool for League of Legends with fuzzy string matching, nicknames, multiple locales, "
    "automatic updates, and translation.",
    long_description=long_description,
    long_description_content_type="text/markdown",
)
